# frozen_string_literal: true

# Copyright from 2019 Paulo Arruda (parrudaj at gmail dot com)
#
# Licensed under the Commons Clause + Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'msgpack'
require 'socket'
require 'open3'
require 'gotta/common/constants'
require 'gotta/run/response'
require 'gotta/run/errors'

module Gotta
  module Run
    # This class is responsible for all dynamic runners.
    # It was designed to be used as a inheritable class.
    # For example:
    # `class RubyRunner < Gotta::Run::DynamicRunner; end`
    class DynamicRunner
      attr_reader :input_queue

      # The initializer takes an input queue, where it should
      # go fetch payloads to send to runners. The default queue
      # is `Gotta::Queue::Memory::Driver`.
      # It also takes an optional argument `runtime_path` that
      # is the path to the runtime script for the language you
      # want to run.
      # You can optionally pass a block in case you need to
      # set extra configuration without having to override the
      # initializer.
      # It will raise `RuntimeScriptNotFound` if `runtime_path`
      # doesn't exist or is not a file.
      def initialize(input_queue: Gotta::Queue.new('input'),
                     runtime_path: nil,
                     &block)
        @input_queue = input_queue
        @runtime_path = runtime_path
        raise RuntimeScriptNotFound, runtime_path unless File.file?(runtime_path)

        @parent_socket, @child_socket = UNIXSocket.pair
        @child_output_r, @child_output_w = IO.pipe
        instance_eval(&block) if block_given?
      end

      def start
        start_runtime
        start_processing
      end

      private

      # Sets the environment variables for the runtime process
      def env
        ENV.to_h.merge(
          'SOCKET_FD' => @child_socket.fileno.to_s,
          'MAX_EXECUTION_TIME' => MAX_EXECUTION_TIME.to_s
        )
      end

      def cmd
        cmd = [
          'ruby',
          '-E',
          'UTF-8:UTF-8',
          @runtime_path
        ].join(' ')
      end

      # Start the runtime from `path`.
      # It runs in a loop to make sure the runner
      # process gets restarted in case it dies.
      def start_runtime
        Thread.new do
          loop do
            spawn_process(cmd)
            sleep 1
          end
        end
      end

      def spawn_process(cmd)
        Open3.popen2(env, cmd, @child_socket.fileno => @child_socket.fileno) do |_stdin, stdout, _status_thread|
          stdout.each_line do |line|
            puts line
          end
        end
      end

      # This method is an infinite loop that pops
      # the last element from the queue and passes
      # the payload to the child process.
      # Then while it waits for the child process to
      # perform its task, this method will setup the
      # response queue.
      def start_processing
        input_queue.process do |payload|
          @parent_socket.send(payload, 0)
          send_response_to(response_queue_for(payload))
        end
      end

      # Sends the response to the response queue.
      # A timeout is used here, but ideally this should be
      # done using socket timeouts.
      # TODO: use socket timeouts instead.
      def send_response_to(response_queue)
        Timeout.timeout(MAX_EXECUTION_TIME) do
          # puts 'Waiting for response from child'
          resp = @parent_socket.recv(MAX_RESPONSE_MSG_SIZE)
          # puts "Got response! #{MessagePack.unpack(resp)}"
          response_queue.push(
            resp
          )
        end
      rescue Timeout::Error
        response_queue.push(Response.timeout.to_msgpack)
      end

      # Initializes the response queue from the name
      # that comes with the payload.
      def response_queue_for(payload)
        object = MessagePack.unpack(payload)
        Gotta::Queue.new(object['response_queue_name'])
      end
    end
  end
end
